<?php

namespace Drupal\xsl_formatter\Utility;

use DOMDocument;
use Drupal\Core\Link;
use Drupal\Core\Url;
use Drupal\xsl_formatter\Exception\XslFormatterException;
use XSLTProcessor;

/**
 * Xml utilities for XSL formatter.
 *
 * Ultra-paranoid and layered with pessimism.
 * Because XML always goes wrong.
 *
 * @package Drupal\xsl_formatter\Utility
 */
class XmlUtilities {

  /**
   * Return a list of xsl files, as found in the search locations.
   */
  public static function enumerateXsls() {
    $modulePath = \Drupal::service('module_handler')->getModule('xsl_formatter')->getPath();
    $paths = [
      'public://xsl' => 'public://xsl',
      'module://xsl_formatter/xsl' => $modulePath . '/xsl',
    ];
    $found = [];
    foreach ($paths as $label => $path) {
      $files = \Drupal::service('file_system')->scanDirectory($path, '/.*\.xsl[t]?/');
      foreach ($files as $file) {
        $found[$file->uri] = $label . '/' . $file->filename;
      }
    }
    return $found;
  }

  /**
   * Find and initialize the transformation template.
   *
   * LOTS of error checking.
   *
   * Allows you to define the path relative to the module, the site,
   * or the files dir.
   *
   * Includes caching retrieval for a bit of speed-up over bulks.
   * XSL is expensive, so if we find ourselves doing it more than once,
   * the parsed file is retained for next time.
   *
   * @param string $xml_file
   *   Path to the XML source file, local absolute path or file wrapper URI.
   * @param string $description
   *   Just used for logging.
   *
   * @throws \Exception
   *   If anything goes wrong (file-not-found or invalid XML).
   *
   * @return \DOMDocument
   *   Parsed Dom Object to work on.
   */
  public static function getXmlDoc($xml_file, $description = "XML file") {
    // Check cache first-off.
    static $xmldocs;
    if (isset($xmldocs[$xml_file])) {
      return $xmldocs[$xml_file];
    }
    if (empty($xml_file)) {
      throw new XslFormatterException("Null $description? Cannot proceed");
    }

    $xml_filepath = $xml_file;
    if (!is_file($xml_filepath)) {
      $xml_filepath = 'public://' . $xml_file;
    }
    if (!is_file($xml_filepath)) {
      $moduleHandler = \Drupal::service('module_handler');
      $xml_filepath = $moduleHandler->getModule('xsl_formatter')->getPath() .
        "/" . $xml_file;
    }

    if (is_file($xml_filepath)) {
      $xml_doc = new DOMDocument();
      $isLoaded = $xml_doc->load($xml_filepath);
      if (!$isLoaded) {
        $xmldocs[$xml_file] = FALSE;
        throw new XslFormatterException("Unable to parse the $description '$xml_filepath' ");
      }

      $xml_docs[$xml_file] = $xml_doc;
    }
    else {
      $xmldocs[$xml_file] = FALSE;
      throw new XslFormatterException("Unable to locate the $description '$xml_file' ");
    }

    return $xml_doc;
  }

  /**
   * Do the actual conversion between XML+XSL.
   *
   * Input and output are full DOM objects in PHP5
   * We return the result STRING, as that's what
   * the process gives us :-/
   * Need to parse it back in again for pipelining.
   *
   * Support for PHP4 XSL removed.
   *
   * If it uses includes, the xsl must have
   * had its documentURI set correctly prior to this, but it can be set in the
   * parameters also.
   *
   * @param \DOMDocument $xml_doc
   *   XML dom document.
   * @param \DOMDocument $xsl_doc
   *   XSL dom document.
   * @param array $parameters
   *   To be passed into the xslt_process().
   *
   * @returns string The result.
   */
  public static function xmldocPlusXsldoc(\DOMDocument $xml_doc,
                                   \DOMDocument $xsl_doc,
                                   array $parameters = []) {
    if (!extension_loaded('xsl')) {
      \Drupal::messenger()->addError(t('PHP XSL library is not enabled. Please see @link1 or @link2',
        [
          '@link1' => Link::fromTextAndUrl('XSL Formatter project on drupal.org', Url::fromUri('https://drupal.org/project/xsl_formatter'))->toString(),
          '@link2' => Link::fromTextAndUrl('XSL installation on php.net', Url::fromUri('http://php.net/manual/en/xsl.installation.php'))->toString(),
        ]));
      return t('PHP XSL library is not enabled. See log for details.');
    }
    $xsltproc = new XSLTProcessor();

    // Attach the xsl rules.
    $xsltproc->importStyleSheet($xsl_doc);
    // Set any processing parameters and flags.
    if ($parameters) {
      foreach ($parameters as $param => $value) {
        $xsltproc->setParameter("", $param, $value);
      }
    }

    $out = $xsltproc->transformToXml($xml_doc);

    if (function_exists('charset_decode_utf_8')) {
      // I just CAN'T trust XML not to have squashed the entities into
      // bytecodes. Expand them before returning or I can never trust that my
      // result here is actually valid to put in anywhere else again.
      return charset_decode_utf_8($out);
    }

    return $out;
  }

}
